(:~
 : This module creates the single edition view. According to the GET parameter
 : "edition" this module gets from edition.xqm all common text lines and the ones
 : that fit a tei:rdg are selected.
 :
 : The whole module is called by xml-utils.xqm.
 :
 : @author Hannes Riebl
 : @author Michelle Weidling
 :)

xquery version "3.1";

module namespace single="http://bdn-edition.de/ns/single";

declare namespace bdn="http://www.bdn-edition.de/ns/1.0";
declare namespace tei="http://www.tei-c.org/ns/1.0";
declare namespace xhtml="http://www.w3.org/1999/xhtml";

import module namespace commons="http://bdn-edition.de/ns/commons" at "commons.xqm";
import module namespace config="https://sade.textgrid.de/ns/config" at "../config.xqm";
import module namespace functx = "http://www.functx.com";
import module namespace id-file="https://bdn-edition.de/ns/id-file" at "id-file.xqm";

declare variable $single:id-map :=
  map:merge
  (
    for $id in doc($config:aux || "id-file.xml")//bdn:id
        return map:entry($id/string(), $id/parent::bdn:file/@uri/string())
  );


declare function single:transform($nodes as node()*, $mode as xs:string*,
$file as xs:string) as node()*
{
  for $node in $nodes
    return
    ( if ($node/@xml:id) then
        commons:make-text-anchor(string($node/@xml:id))
      else
        (),
      try
      { typeswitch($node)
        case text() return
          let $text := text{replace($node, "@", " ")}
          return
    (:          let $app := $node/following-sibling::*[1][self::tei:app]:)
    (:          let $next-text := $app/following-sibling::node()[1][self::text()]:)
    (:          let $next-text-first-char := substring($next-text, 1, 1):)
    (:          return:)
    (:              (: example: Griesbach 250bg, #a "einen neuen Zuwachs, wenn man":):)
    (:              if ($app/tei:rdg[matches(@wit, $mode) and @type = "om"]:)
    (:              and matches($next-text-first-char, "[,\.\?!\)]")) then:)
    (:                  text{functx:right-trim($text)}:)
    (:              else:)
                      $text

        case comment() return
          ()

        case element(tei:teiHeader) return
          ()

        case element(tei:div) return
          switch($node/@type)
          case ""
            return
            element xhtml:div
            {
              single:transform($node/node(), $mode, $file)
            }
          case "synopsis" return
            (
              single:transform($node/tei:head, $mode, $file),
              element xhtml:table
              {
                element xhtml:tr
                {
                  element xhtml:td
                  {
                    single:transform($node/descendant::tei:div[@type = "letter"][1]/tei:head, $mode, $file)
                  },
                  element xhtml:td
                  {
                    single:transform($node/descendant::tei:div[@type = "letter"][2]/tei:head, $mode, $file)
                  }
                },
                (: ATTENTION: this implies that the two div[@type = 'letter']
                always have the same number of ab :)
                (
                  for $ab in $node/descendant::tei:div[@type = "letter"][1]/tei:ab
                  return
                    element xhtml:tr
                    {
                      element xhtml:td
                      {
                        attribute class {"syn-td"},
                        single:transform($node/node(), $mode, $file)
                      },
                      element xhtml:td
                      {
                        attribute class {"syn-td"},
                        single:transform($node/descendant::div[@type = 'letter'][2]/ab[@xml:id = $ab/@corresp], $mode, $file)
                      }
                    }
                )
              }
            )
          case "letter" return
            if ($node/ancestor::tei:div[@type = "synopsis"]) then
              ()
            else
              single:transform($node/node(), $mode, $file)
          default return
            element xhtml:div
            {
              attribute class {$node/@type},
              single:transform($node/node(), $mode, $file)
            }

        case element(tei:head) return
          switch($node/@type)
          case "main" return
            element xhtml:div
            {
              attribute class {"list-main-head"},
              single:transform($node/node(), $mode, $file)
            }
          case "sub" return
            element xhtml:div
            {
              attribute class {"list-sub-head"},
              single:transform($node/node(), $mode, $file)
            }
          default return
            element xhtml:h2
            {
              if ($node/@copyOf) then
                single:apply-corresp-node($node/@copy-of, $mode, $file, false())
              else
                single:transform($node/node(), $mode, $file)
            }

        case element(tei:seg) return
          if ($node/@copyOf) then
            single:apply-corresp-node($node/@copyOf, $mode, $file, false())
          else if ($node/@type = "margin") then
                element xhtml:span
                {
                    attribute class {"margin"},
                    text{"{"},
                    single:transform($node/node(), $mode, $file),
                    text{"}"}
                }
          else
            single:transform($node/node(), $mode, $file)


        case element(tei:front) return
          element xhtml:div
          {
            attribute class {"preface"},
            single:transform($node/node(), $mode, $file)
          }

        case element(tei:titlePart) return
          switch($node/@type)
          case "main" return
            element xhtml:h1
            {
              attribute class {"main-title"},
              single:transform($node/node(), $mode, $file)
            }
          case "sub" return
            element xhtml:h2
            {
              attribute class {"sub-title"},
              single:transform($node/node(), $mode, $file)
            }
          case "volume" return
            element xhtml:div
            {
              attribute class {"volume"},
              single:transform($node/node(), $mode, $file)
            }
          default return ()

        case element(tei:app) return
          (if ($node/child::*[matches(@wit, $mode)]) then
            single:transform($node/child::*[matches(@wit, $mode)], $mode, $file)
          else
            single:transform($node/child::tei:lem, $mode, $file),
          if (single:is-next-text-in-app($node)
          and single:is-next-app-omitted($node, $mode)) then
            ()
          else
            commons:make-whitespace-after($node))

        case element(tei:rdg) return
        single:transform($node/node(), $mode, $file)

        case element(tei:rdgMarker) return
          ()

        case element(tei:foreign) return
          if ($node[@xml:lang = "grc" and descendant::tei:hi]) then
            element xhtml:span
            {
              attribute class {"greek"},
              single:transform($node/node(), $mode, $file)
            }
          else
            element xhtml:span
            {
              attribute class {"foreign"},
              single:transform($node/node(), $mode, $file)
            }

        case element(tei:aligned) return
          element xhtml:div
          {
            attribute class {$node/@rend},
            single:transform($node/node(), $mode, $file)
          }

        case element(tei:hi) return
          if ($node/@rend) then
            element xhtml:span
            {
              if ($node/@rend = "spaced-out"
              and matches(substring($node/following::text()[1], 1, 1), "[,!\?\.:;]")) then
                  attribute class {"spaced-out spaced-out-no-right-padding"}
              else
                  attribute class {$node/@rend},
              single:transform($node/node(), $mode, $file)
            }
          else
            element xhtml:span
            {
              attribute class {"hi"},
              single:transform($node/node(), $mode, $file)
            }

        case element(tei:index) return
          ()

        case element(tei:label) return
          if ($node[@type = "headword"]) then
            element xhtml:b
            {
              single:transform($node/node(), $mode, $file)
            }
          else
            element xhtml:span
            {
              attribute class {"label-wrapper"},
              single:transform($node/node(), $mode, $file)
            }

        case element(tei:list) return
          element xhtml:ul
          {
            (if ($node/parent::tei:item) then
              attribute class {"nested-list"}
            else if ($node/ancestor::tei:rdg) then
              attribute class {"rdg-list"}
            else
              attribute class {"text-list"}),
            single:transform($node/node(), $mode, $file)
          }


        case element(tei:l) return
          element xhtml:span
          {
            attribute class {"line"},
            single:transform($node/node(), $mode, $file)
          }

        case element(tei:lg) return
          element xhtml:div
          {
            attribute class {"linegroup"},
            single:transform($node/node(), $mode, $file)
          }

        case element(tei:milestone) return
            let $edition := replace(replace($node/@edRef, "#", ""), " ", "")
            return
              if ($node/@unit = "fn-break"
              and $edition = $mode) then
                element xhtml:span
                {
                  attribute class {"fn-break"},
                  string("|" || $edition || $node/@n || "|")
                }
              else
                  ()

        case element(tei:note) return
          if ($node/@type = "editorial-commentary") then
            ()
          else
            (if ($node/ancestor::tei:p) then
              <br />
            else
              (),
            element xhtml:div
            {
                attribute class {"note"},
                single:transform($node/node(), $mode, $file)
            })

        case element(tei:ptr) return
          ()

        case element(tei:titlePage) return
          element xhtml:span
          {
            attribute class {"title-page"},
            single:transform($node/node(), $mode, $file)
          }

        case element(tei:ab) return
          if ($node[@type = "half-title"]) then
            element xhtml:div
            {
              attribute class {$node/@type},
              single:transform($node/node(), $mode, $file)
            }
          else
            element xhtml:span
            {
              attribute class {"ab"},
              single:transform($node/node(), $mode, $file)
            }

        case element(tei:cb) return
          element xhtml:span
          {
            attribute class {"cb"},
            string("|" || replace($node/@edRef, '[# ]+', '') || $node/@n || "|")
          }

        case element(tei:pb) return
          if (matches($node/@edRef, $mode)) then
            (
              commons:make-whitespace-before($node),
              element xhtml:span
              {
                attribute class {"pb"},
                concat
                (
                  string("|" || replace($node/@edRef, '[# ]+', '')),
                  (if ($node/@type = 'sp') then string("[") else ""),
                  string($node/@n),
                  (if ($node/@type = 'sp') then string("]") else ""),
                  string("|")
                )
              }
            )
          else if (not($node/following-sibling::*[1][self::tei:pb])) then
            commons:make-whitespace-after($node)
          else
            ()


        case element(tei:back) return
          single:transform($node/node(), $mode, $file)

        case element(tei:item) return
          element xhtml:li
          {
            if ($node/@xml:id) then
              attribute id {$node/@xml:id}
            else
              (),
            if ($node/ancestor::tei:div[@type = "contents"])
            then
              attribute class {"list-item-toc"}
            else
              attribute class {"list-item"},

            if ($node/@copyOf) then
              single:apply-corresp-node($node/@copyOf, $mode, $file, false())
            else
              single:transform($node/node(), $mode, $file)
          }


        case element(tei:choice) return
          if ($node//tei:abbr) then
            single:transform($node//tei:abbr, $mode, $file)

          else if ($node//tei:orig) then
            single:transform($node//tei:orig, $mode, $file)

          (: compositor's errors :)
          else if ($node/tei:corr
          and ($node/..[@type = "typo-correction"][contains(@wit, $mode)]
            or not($node/parent::tei:lem/following-sibling::tei:rdg[contains(@wit, $mode)]))) then
            (single:transform($node/tei:corr, $mode, $file),
            single:make-typo-corr($node, $mode, $file),
            (: compositor's error with authorial correction :)
            if ($node/tei:corr[@type = "authorial"]) then
              element xhtml:a
              {
                attribute href {commons:get-link-url($node, $mode)},
                <i class="fas fa-pen-fancy"></i>
              }
            else
              ()
            )

          (: authorial corrigenda :)
          else if ($node/tei:corr[@type = "authorial"]) then
            (single:transform($node//tei:corr, $mode, $file),
            element xhtml:a
            {
              attribute href {commons:get-link-url($node, $mode)},
              string("@TEST@"),
              <i class="fas fa-pen-fancy"></i>
            })
          else
            () (:TODO: Satzfehler :)

        case element(tei:join) return
          switch($node/@result)
          case "p" return
            element xhtml:div
            {
              attribute class {"p"},
              single:apply-join($node, $mode, $file)
            }
          case "item" return
            element xhtml:li
            {
              attribute class {"list-item"},
              single:apply-join($node, $mode, $file)
            }
          (: the default case is that join has no @result :)
          default return
            single:apply-join($node, $mode, $file)


        case element(tei:p) return
          (: since we have xhtml:div nested in p, e.g. for compositor's
          errors, we need to choose xhtml:div here instead of xhtml:p.
          otherwise the HTML content model is violated. :)
          element xhtml:div
          {
            if ($node/@rend= "center-aligned") then
                attribute class {"center-aligned"}
            else
                attribute class {"p"},
            single:transform($node/node(), $mode, $file),
            (:if (preceding-sibling::p
            or (ancestor::app[@type = 'structural-variance']/parent::div[@type = 'section']
            and ancestor::div[@type = 'section']/child::*[2] != ancestor::app[@type = 'structural-variance']):)
            (: in case a paragraph starts directly with an app[@type = 'structural-variance'] :)
            (:or ancestor::app[@type = 'structural-variance']/parent::div[@type = 'section']
              and not(//*[@xml:id = $node/@copyID]/ancestor::app[@type = 'structural-variance'])
            or ancestor::app/preceding-sibling::*[1][self::p]
            or preceding-sibling::*[1][self::list]
            or preceding-sibling::*[1]/child::*[last()][self::list]
            or preceding-sibling::*[1][self::app]/lem/child::*[last()][self::list])
              then
              attribute{"style"}{"text-indent: 15px;"}
              else
              attribute{"style"}{"margin-left: 0;"},:)
            if ($node/@copyOf) then
              single:apply-corresp-node($node/@copyOf, $mode, $file, false())
            else
              single:transform($node/node(), $mode, $file)
            }


        case element(tei:ref) return
          (
            element xhtml:span
            {
              attribute class {"ref"},
              single:transform($node/node(), $mode, $file),
              string(" "),
              element xhtml:a
              {
                attribute class {"text-link"},
                attribute href {commons:get-link-url($node, $mode)},
                <i class="fas fa-external-link-alt"></i>
              }
            },
            if (commons:starts-next-text-with-punctuation($node, $mode)) then
              ()
            else
              text{" "}
          )

        case element(tei:lb) return
          if ($node/preceding-sibling::*
            [ self::tei:titlePart
              or self::tei:docImprint
              or self::tei:head
            ]
          )
          then
            ()
          else
            <br />

        case element(tei:docImprint) return
            (single:transform($node/node(), $mode, $file),
            <br />)

        case element(tei:supplied) return
          (
            text {"["},
            single:transform($node/node(), $mode, $file),
            text{"]"}
          )

        case element (tei:floatingText) return
            element xhtml:div
            {
                attribute class {"floating-text"},
                single:transform($node/node(), $mode, $file)
            }


        case element (tei:signed) return
            element xhtml:div
            {
                attribute class {"signed"},
                single:transform($node/node(), $mode, $file)
            }

        default return
          single:transform($node/node(), $mode, $file),
          commons:make-whitespace-after($node)
      } catch *
      { element xhtml:p
          {
              attribute class {"error"},
              text{"Error: "}, $err:code,
              text{"Description: "}, $err:description,
              text{"Node: "}, $node/name()
          }
      }
    )
};

(:~
 : Finds and applies the node that is referenced with @copyOf.
 : In case of tei:join we have to apply only the child nodes of the corresponding
 : nodes because tei:join defines a new resulting element in @result that makes
 : the original nodes obsolete.
 :
 : @author Michelle Weidling
 : @version 1.0
 : @param $copyOf A reference to another node
 : @param $mode The current edition
 : @param $file
 : @param $join A flag indicating if the calling element is a tei:join
 : @return The transformed node the @copyOf refers to
 :)
declare function single:apply-corresp-node($copyOf as xs:string, $mode as
xs:string, $file as xs:string, $apply-children as xs:boolean)
as node()*
{
  let $corresp-node := single:find-corresp-node($copyOf, $file)
  return
      if ($apply-children) then
          for $child in $corresp-node return
              single:transform($child, $mode, $file)
      else
            single:transform($corresp-node, $mode, $file)
};


(:~
 : Retrieves the target node of a given @copyOf-reference. Since references can
 : refer to another chapter (or even another volume) of a BdN work, the target
 : is looked up in the ID file.
 :
 : @author Michelle Weidling
 : @version 1.0
 : @param $copyOf A reference to another node
 : @param $file The current file
 : @return The node the @copyOf refers to
 :)
declare function single:find-corresp-node($copyOf as xs:string, $file as
xs:string) as node()?
{
  let $author := commons:get-author($file)
  let $copyOf := replace($copyOf, "#", "")
  let $target-file := map:get($single:id-map, $copyOf)
  let $doc := doc($config:edition || $author || "/xml/" || $target-file || ".xml")
  return
    $doc//*[@xml:id = $copyOf]
};


(:~
 : Creates the icon for a compositor's error and initiates the construction of
 : the respective modal div.
 :
 : @author Michelle Weidling
 : @version 1.0
 : @param $node The the current tei:choice
 : @param $mode The current edition
 : @return The modal div, a whitespace and the icon
 :)
declare function single:make-typo-corr($node as element(tei:choice),
$mode as xs:string, $file as xs:string) as node()*
{
  let $random-id := single:make-random-id()
  let $id := util:hash($random-id, "md5")
  let $edition :=
    if ($node/parent::tei:lem[not(@wit)]) then
      (: single file names don't have a hyphen, only the ones that are the
      result of semantic splitting :)
      if (matches($file, "-")) then
        let $author := substring-before($file, "-")
        let $full-file := doc($config:full-xmls || $author || "-full.xml")
        return $full-file//tei:witness[@n = "base-text"]/@xml:id/string()
      else
        $node/ancestor::tei:TEI//tei:witness[@n = "base-text"]/@xml:id
    else
      $mode
  return
    (single:make-typo-corr-modal($node, $edition, $mode, $id, $file),
    text{" "},
    <a class="comp-error" data-target="#{$id}">
      <i class="fa fa-wrench" aria-hidden="true" data-toggle="modal" />
    </a>)
};


(:~
 : Creates the modal div for a compositor's error.
 :
 : @author Michelle Weidling
 : @version 1.0
 : @param $node The the current tei:choice
 : @param $edition The edition in which an error occurs
 : @param $mode The current edition
 : @param $id The target id
 : @return The modal div
 :)
declare function single:make-typo-corr-modal($node as element(tei:choice),
$edition as xs:string, $mode as xs:string, $id as xs:string, $file as xs:string)
as element(xhtml:div)
{
  <div class="modal fade" role="dialog" id="{$id}">
    <div class="modal-dialog modal-typo">
      <div class="modal-content">
        <div class="modal-header">
          <button type="button" class="close" data-dismiss="modal">
            × <!-- times -->
          </button>
          {
            if ($node/tei:corr/@type = "authorial") then
              let $sic :=
                if ($node/tei:sic[not(child::node() or child::node())]) then
                  <em>&#248;</em>
                else
                  single:transform($node/tei:sic, $mode, $file)
              return
                (string("Autorschaftliche Korrektur von: "),
                $sic,
                string(" (" || $edition || ")"))
            else
              let $sic :=
                if ($node/tei:sic[not(child::* or child::node())]) then
                  <em>&#248;</em>
                else
                  single:transform($node/tei:sic, $mode, $file)
              return
                (string("Editorische Korrektur von: "),
                $sic,
                string(" (" || $edition || ")"))
          }
        </div>
      </div>
    </div>
  </div>
};


declare function single:is-next-text-in-app($node as node()*) as xs:boolean
{
  if ($node/following::text()[1][normalize-space(.) != ""][ancestor::tei:app])then
    true()
  else
    false()
};

declare function single:is-next-app-omitted($node as node()*, $mode as xs:string)
as xs:boolean
{
  let $next-app := $node/following-sibling::*[1][self::tei:app]
  return
    if ($next-app/child::tei:rdg[matches(@wit, $mode) and @type = "om"]) then
      true()
    else
      false()
};

(:~
 : In cases like tei:join several references are stated in a @target. This
 : function returns all corresponding nodes.
 :
 : @author Michelle Weidling
 : @version 1.0
 : @param $target The @target of a e.g. tei:join
 : @param $mode The current edition
 : @return The nodes the @target refers to
 :)
declare function single:apply-join($node as element(tei:join), $mode as xs:string,
$file as xs:string) as node()*
{
    let $apply-only-children :=
        if ($node/@scope = "root") then
            false()
        else
            true()
    return
        for $copyOf in tokenize($node/@target, " ")
            let $applied-node := single:apply-corresp-node($copyOf, $mode, $file, $apply-only-children)
            let $first-text := $applied-node[1]/descendant-or-self::text()[1]
            let $conjunction :=
              if (starts-with($first-text, ",")) then
                text{""}
              else
                text{" "}
            return ($conjunction, $applied-node)
};

(:~
 : Creates a random ID of 40 letters consisting of Basic Latin letters and
 : numbers. This function is needed since the usage of generate-id() (which we
 : originally used for the linking between icons and a modal dialogue) lead to
 : ID duplettes.
 :)
declare function single:make-random-id() as xs:string
{
    let $charset := "ABCDEFGHIJKLMNOPQRSTUVXYZabcdefghijklmnopqrstuvwxyz0123456789"
    let $id :=
        for $iii in 1 to 40 return
            let $number := util:random(40)
            return
                substring($charset, $number, 1)
    return string-join($id, "")
};
